@@ -247,8 +247,7 @@ modsem_pi <- function(model.syntax = NULL,
     return(est)
   }
 
-  if (!is.data.frame(data))
-    data <- as.data.frame(data)
+  data <- as.data.frame(data)
 
   if (rcs) { # use reliability-correct single items?
     if (!is.null(rcs.choose))
@@ -382,7 +381,12 @@ modsem_pi <- function(model.syntax = NULL,
     modelSpec$coefParTable <- coefParTable
   }
 
-  structure(modelSpec, class = c("modsem_pi", "modsem"), method = method)
+  structure(
+    modelSpec,
+    class = c("modsem_pi", "modsem"),
+    method = method,
+    isRCS_Model = rcs
+  )
 }
 
 
@@ -781,6 +785,7 @@ modsemPICluster <- function(model.syntax = NULL,
   stopif(length(syntaxBlocks) != length(levelHeaders), "Different number of blocks than level headers!")
 
   data$ROW_IDENTIFIER_ <- seq_len(nrow(data))
+  has.interaction      <- FALSE
   newSyntax <- ""
   newData <- NULL
 
@@ -788,7 +793,11 @@ modsemPICluster <- function(model.syntax = NULL,
     syntaxBlock <- syntaxBlocks[[i]]
     levelHeader <- levelHeaders[[i]]
 
-    if (!NROW(modsemify(syntaxBlock))) next
+    parsedBlock <- modsemify(syntaxBlock)
+    if (!NROW(parsedBlock)) next
+
+    if (any(grepl(":", parsedBlock$rhs)))
+        has.interaction <- TRUE
 
     newBlockSyntax <- get_pi_syntax(
       model.syntax = syntaxBlock,
@@ -863,7 +872,9 @@ modsemPICluster <- function(model.syntax = NULL,
     )
   }
 
-  modelSpec <- list(syntax = newSyntax, data = newData)
+  modelSpec <- list(syntax = newSyntax, data = newData,
+                    has.interaction = has.interaction)
+
   if (run) {
     lavWrapper <- getWarningWrapper(silent = suppress.warnings.lavaan)
     lavEst <- tryCatch({
